@@ -11,16 +11,19 @@
 # under the License.
 
 import json
-import mock
 import re
 
+import mock
 from testtools import matchers
 
+from keystoneauth1 import adapter
 from keystoneauth1 import discover
 from keystoneauth1 import exceptions
 from keystoneauth1 import fixture
+from keystoneauth1 import noauth
 from keystoneauth1 import session
 from keystoneauth1.tests.unit import utils
+from keystoneauth1 import token_endpoint
 
 
 BASE_HOST = 'http://keystone.example.com'
@@ -130,7 +133,7 @@ V3_AUTH_RESPONSE = json.dumps({
 CINDER_EXAMPLES = {
     "versions": [
         {
-            "status": "CURRENT",
+            "status": discover.Status.CURRENT,
             "updated": "2012-01-04T11:33:21Z",
             "id": "v1.0",
             "links": [
@@ -141,7 +144,7 @@ CINDER_EXAMPLES = {
             ]
         },
         {
-            "status": "CURRENT",
+            "status": discover.Status.CURRENT,
             "updated": "2012-11-21T11:33:21Z",
             "id": "v2.0",
             "links": [
@@ -152,7 +155,7 @@ CINDER_EXAMPLES = {
             ]
         },
         {
-            "status": "CURRENT",
+            "status": discover.Status.CURRENT,
             "updated": "2012-11-21T11:33:21Z",
             "id": "v3.0",
             "version": "3.27",
@@ -176,7 +179,7 @@ CINDER_EXAMPLES = {
 GLANCE_EXAMPLES = {
     "versions": [
         {
-            "status": "CURRENT",
+            "status": discover.Status.CURRENT,
             "id": "v2.2",
             "links": [
                 {
@@ -186,7 +189,7 @@ GLANCE_EXAMPLES = {
             ]
         },
         {
-            "status": "SUPPORTED",
+            "status": discover.Status.SUPPORTED,
             "id": "v2.1",
             "links": [
                 {
@@ -196,7 +199,7 @@ GLANCE_EXAMPLES = {
             ]
         },
         {
-            "status": "SUPPORTED",
+            "status": discover.Status.SUPPORTED,
             "id": "v2.0",
             "links": [
                 {
@@ -206,7 +209,7 @@ GLANCE_EXAMPLES = {
             ]
         },
         {
-            "status": "CURRENT",
+            "status": discover.Status.CURRENT,
             "id": "v1.1",
             "links": [
                 {
@@ -216,7 +219,7 @@ GLANCE_EXAMPLES = {
             ]
         },
         {
-            "status": "SUPPORTED",
+            "status": discover.Status.SUPPORTED,
             "id": "v1.0",
             "links": [
                 {
@@ -283,7 +286,7 @@ class CatalogHackTests(utils.TestCase):
 class DiscoverUtils(utils.TestCase):
 
     def test_version_number(self):
-        def assertVersion(inp, out):
+        def assertVersion(out, inp):
             self.assertEqual(out, discover.normalize_version_number(inp))
 
         def versionRaises(inp):
@@ -291,24 +294,24 @@ class DiscoverUtils(utils.TestCase):
                               discover.normalize_version_number,
                               inp)
 
-        assertVersion('v1.2', (1, 2))
-        assertVersion('v11', (11, 0))
-        assertVersion('1.2', (1, 2))
-        assertVersion('1.5.1', (1, 5, 1))
-        assertVersion('1', (1, 0))
-        assertVersion(1, (1, 0))
-        assertVersion(5.2, (5, 2))
-        assertVersion('3.20', (3, 20))
+        assertVersion((1, 2), 'v1.2')
+        assertVersion((11, 0), 'v11')
+        assertVersion((1, 2), '1.2')
+        assertVersion((1, 5, 1), '1.5.1')
+        assertVersion((1, 0), '1')
+        assertVersion((1, 0), 1)
+        assertVersion((5, 2), 5.2)
+        assertVersion((3, 20), '3.20')
         assertVersion((6, 1), (6, 1))
-        assertVersion([1, 40], (1, 40))
-        assertVersion((1,), (1, 0))
-        assertVersion(['1'], (1, 0))
-        assertVersion('latest', (discover.LATEST, discover.LATEST))
-        assertVersion(['latest'], (discover.LATEST, discover.LATEST))
-        assertVersion(discover.LATEST, (discover.LATEST, discover.LATEST))
-        assertVersion((discover.LATEST,), (discover.LATEST, discover.LATEST))
-        assertVersion('10.latest', (10, discover.LATEST))
-        assertVersion((10, 'latest'), (10, discover.LATEST))
+        assertVersion((1, 40), [1, 40])
+        assertVersion((1, 0), (1,))
+        assertVersion((1, 0), ['1'])
+        assertVersion((discover.LATEST, discover.LATEST), 'latest')
+        assertVersion((discover.LATEST, discover.LATEST), ['latest'])
+        assertVersion((discover.LATEST, discover.LATEST), discover.LATEST)
+        assertVersion((discover.LATEST, discover.LATEST), (discover.LATEST,))
+        assertVersion((10, discover.LATEST), '10.latest')
+        assertVersion((10, discover.LATEST), (10, 'latest'))
         assertVersion((10, discover.LATEST), (10, discover.LATEST))
 
         versionRaises(None)
@@ -320,96 +323,138 @@ class DiscoverUtils(utils.TestCase):
 
     def test_version_args(self):
         """Validate _normalize_version_args."""
-        def assert_min_max(in_ver, in_min, in_max, out_min, out_max):
+        def assert_min_max(in_ver, in_min, in_max, in_type, out_min, out_max):
             self.assertEqual(
                 (out_min, out_max),
-                discover._normalize_version_args(in_ver, in_min, in_max))
+                discover._normalize_version_args(
+                    in_ver, in_min, in_max, service_type=in_type))
 
-        def normalize_raises(ver, min, max):
-            self.assertRaises(ValueError,
-                              discover._normalize_version_args, ver, min, max)
+        def normalize_raises(ver, min, max, in_type):
+            self.assertRaises(
+                ValueError,
+                discover._normalize_version_args,
+                ver, min, max, service_type=in_type)
 
-        assert_min_max(None, None, None,
+        assert_min_max(None, None, None, None,
                        None, None)
-        assert_min_max(None, None, 'v1.2',
+        assert_min_max(None, None, 'v1.2', None,
                        None, (1, 2))
-        assert_min_max(None, 'v1.2', 'latest',
+        assert_min_max(None, 'v1.2', 'latest', None,
                        (1, 2), (discover.LATEST, discover.LATEST))
-        assert_min_max(None, 'v1.2', '1.6',
+        assert_min_max(None, 'v1.2', '1.6', None,
                        (1, 2), (1, 6))
-        assert_min_max(None, 'v1.2', '1.latest',
+        assert_min_max(None, 'v1.2', '1.latest', None,
                        (1, 2), (1, discover.LATEST))
-        assert_min_max(None, 'latest', 'latest',
+        assert_min_max(None, 'latest', 'latest', None,
                        (discover.LATEST, discover.LATEST),
                        (discover.LATEST, discover.LATEST))
-        assert_min_max(None, 'latest', None,
+        assert_min_max(None, 'latest', None, None,
                        (discover.LATEST, discover.LATEST),
                        (discover.LATEST, discover.LATEST))
-        assert_min_max(None, (1, 2), None,
+        assert_min_max(None, (1, 2), None, None,
                        (1, 2), (discover.LATEST, discover.LATEST))
-        assert_min_max('', ('1', '2'), (1, 6),
+        assert_min_max('', ('1', '2'), (1, 6), None,
                        (1, 2), (1, 6))
-        assert_min_max(None, ('1', '2'), (1, discover.LATEST),
+        assert_min_max(None, ('1', '2'), (1, discover.LATEST), None,
                        (1, 2), (1, discover.LATEST))
-        assert_min_max('v1.2', '', None,
+        assert_min_max('v1.2', '', None, None,
                        (1, 2), (1, discover.LATEST))
-        assert_min_max('1.latest', None, '',
+        assert_min_max('1.latest', None, '', None,
                        (1, discover.LATEST), (1, discover.LATEST))
-        assert_min_max('v1', None, None,
+        assert_min_max('v1', None, None, None,
                        (1, 0), (1, discover.LATEST))
-        assert_min_max('latest', None, None,
+        assert_min_max('latest', None, None, None,
                        (discover.LATEST, discover.LATEST),
                        (discover.LATEST, discover.LATEST))
+        assert_min_max(None, None, 'latest', 'volumev2',
+                       (2, 0), (2, discover.LATEST))
+        assert_min_max(None, None, None, 'volumev2',
+                       (2, 0), (2, discover.LATEST))
 
-        normalize_raises('v1', 'v2', None)
-        normalize_raises('v1', None, 'v2')
-        normalize_raises(None, 'latest', 'v1')
-        normalize_raises(None, 'v1.2', 'v1.1')
-        normalize_raises(None, 'v1.2', 1)
+        normalize_raises('v1', 'v2', None, None)
+        normalize_raises('v1', None, 'v2', None)
+        normalize_raises(None, 'latest', 'v1', None)
+        normalize_raises(None, 'v1.2', 'v1.1', None)
+        normalize_raises(None, 'v1.2', 1, None)
+        normalize_raises('v2', None, None, 'volumev3')
+        normalize_raises('v3', None, None, 'volumev2')
+        normalize_raises(None, 'v2', None, 'volumev3')
+        normalize_raises(None, None, 'v3', 'volumev2')
 
     def test_version_to_string(self):
-        def assert_string(inp, out):
+        def assert_string(out, inp):
             self.assertEqual(out, discover.version_to_string(inp))
 
-        assert_string((discover.LATEST,), 'latest')
-        assert_string((discover.LATEST, discover.LATEST), 'latest')
-        assert_string((discover.LATEST, discover.LATEST, discover.LATEST),
-                      'latest')
-        assert_string((1,), '1')
-        assert_string((1, 2), '1.2')
-        assert_string((1, discover.LATEST), '1.latest')
+        assert_string('latest', (discover.LATEST,))
+        assert_string('latest', (discover.LATEST, discover.LATEST))
+        assert_string('latest',
+                      (discover.LATEST, discover.LATEST, discover.LATEST))
+        assert_string('1', (1,))
+        assert_string('1.2', (1, 2))
+        assert_string('1.latest', (1, discover.LATEST))
 
     def test_version_between(self):
         def good(minver, maxver, cand):
-            self.assertTrue(discover._version_between(minver, maxver, cand))
+            self.assertTrue(discover.version_between(minver, maxver, cand))
 
         def bad(minver, maxver, cand):
-            self.assertFalse(discover._version_between(minver, maxver, cand))
+            self.assertFalse(discover.version_between(minver, maxver, cand))
 
-        def exc(minver, maxver, cand):
-            self.assertRaises(ValueError,
-                              discover._version_between, minver, maxver, cand)
+        def exc(excls, minver, maxver, cand):
+            self.assertRaises(excls,
+                              discover.version_between, minver, maxver, cand)
 
+        # candidate required
+        exc(ValueError, (1, 0), (1, 0), None)
+        exc(ValueError, 'v1.0', '1.0', '')
+        # malformed candidate
+        exc(TypeError, None, None, 'bogus')
+        exc(TypeError, None, None, (1, 'two'))
+        # malformed min_version
+        exc(TypeError, 'bogus', None, (1, 0))
+        exc(TypeError, (1, 'two'), None, (1, 0))
+        # malformed max_version
+        exc(TypeError, None, 'bogus', (1, 0))
+        exc(TypeError, None, (1, 'two'), (1, 0))
+
+        # fail on minimum
+        bad((2, 4), None, (1, 55))
+        bad('v2.4', '', '2.3')
+        bad('latest', None, (2, 3000))
+        bad((2, discover.LATEST), '', 'v2.3000')
+        bad((2, 3000), '', (1, discover.LATEST))
+        bad('latest', None, 'v1000.latest')
+        # fail on maximum
+        bad(None, (2, 4), (2, 5))
+        bad('', 'v2.4', '2.5')
+        bad(None, (2, discover.LATEST), (3, 0))
+        bad('', '2000.latest', 'latest')
+
+        # candidate matches a bound
         good((1, 0), (1, 0), (1, 0))
+        good('1.0', '2.9', '1.0')
+        good('v1.0', 'v2.9', 'v2.9')
+        # properly in between
         good((1, 0), (1, 10), (1, 2))
-        good(None, (1, 10), (1, 2))
-        good((1, 20), (2, 0), (1, 21))
-        good((1, 0), (2, discover.LATEST), (1, 21))
-        good((1, 0), (2, discover.LATEST), (1, discover.LATEST))
-        good((1, 50), (2, discover.LATEST), (2, discover.LATEST))
-
-        bad((discover.LATEST, discover.LATEST),
-            (discover.LATEST, discover.LATEST), (1, 0))
-        bad(None, None, (1, 0))
-        bad((1, 50), (2, discover.LATEST), (3, 0))
-        bad((1, 50), (2, discover.LATEST), (3, discover.LATEST))
-        bad((1, 50), (2, 5), (2, discover.LATEST))
-
-        exc((1, 0), (1, 0), None)
-        exc('v1.0', (1, 0), (1, 0))
-        exc((1, 0), 'v1.0', (1, 0))
-        exc((1, 0), (1, 0), 'v1.0')
-        exc((1, 0), None, (1, 0))
+        good('1', '2', '1.2')
+        # no lower bound
+        good(None, (2, 5), (2, 3))
+        # no upper bound
+        good('2.5', '', '2.6')
+        # no bounds at all
+        good('', '', 'v1')
+        good(None, None, (999, 999))
+        good(None, None, 'latest')
+        # Various good 'latest' scenarios
+        good((discover.LATEST, discover.LATEST),
+             (discover.LATEST, discover.LATEST),
+             (discover.LATEST, discover.LATEST))
+        good((discover.LATEST, discover.LATEST), None,
+             (discover.LATEST, discover.LATEST))
+        good('', 'latest', 'latest')
+        good('2.latest', '3.latest', '3.0')
+        good('2.latest', None, (55, 66))
+        good(None, '3.latest', '3.9999')
 
 
 class VersionDataTests(utils.TestCase):
@@ -446,6 +491,52 @@ class VersionDataTests(utils.TestCase):
 
             self.assertTrue(mock.called_once)
 
+    def test_version_data_override_version_url(self):
+        # if the request url is versioned already, just return it.
+        self.requests_mock.get(
+            V3_URL, status_code=200,
+            json={'version': fixture.V3Discovery('http://override/identity/v3')
+                  }
+        )
+
+        disc = discover.Discover(self.session, V3_URL)
+        version_data = disc.version_data()
+
+        for v in version_data:
+            self.assertEqual(v['version'], (3, 0))
+            self.assertEqual(v['status'], discover.Status.CURRENT)
+            self.assertEqual(v['raw_status'], 'stable')
+            self.assertEqual(v['url'], V3_URL)
+
+        # if the request url is not versioned, just add version info to it.(
+        # do not changed the url's netloc or path)
+        self.requests_mock.get(
+            BASE_URL, status_code=200,
+            json={'version': fixture.V3Discovery('http://override/identity/v3')
+                  }
+        )
+
+        disc = discover.Discover(self.session, BASE_URL)
+        version_data = disc.version_data()
+
+        for v in version_data:
+            self.assertEqual(v['version'], (3, 0))
+            self.assertEqual(v['status'], discover.Status.CURRENT)
+            self.assertEqual(v['raw_status'], 'stable')
+            self.assertEqual(v['url'], V3_URL)
+
+    def test_version_data_unknown(self):
+        discovery_fixture = fixture.V3Discovery(V3_URL)
+        discovery_fixture.status = 'hungry'
+        discovery_doc = _create_single_version(discovery_fixture)
+
+        self.requests_mock.get(V3_URL, status_code=200, json=discovery_doc)
+
+        disc = discover.Discover(self.session, V3_URL)
+        clean_data = disc.version_data(allow_unknown=True)
+
+        self.assertEqual(discover.Status.UNKNOWN, clean_data[0]['status'])
+
     def test_version_data_individual(self):
         mock = self.requests_mock.get(V3_URL,
                                       status_code=200,
@@ -463,11 +554,101 @@ class VersionDataTests(utils.TestCase):
 
         for v in clean_data:
             self.assertEqual(v['version'], (3, 0))
+            self.assertEqual(v['status'], discover.Status.CURRENT)
             self.assertEqual(v['raw_status'], 'stable')
             self.assertEqual(v['url'], V3_URL)
 
         self.assertTrue(mock.called_once)
 
+    def test_version_data_legacy_ironic_no_override(self):
+        """Validate detection of legacy Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}]},
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        plugin = noauth.NoAuth()
+        a = adapter.Adapter(
+            self.session,
+            auth=plugin,
+            service_type='baremetal')
+
+        self.assertIsNone(a.get_api_major_version())
+
+    def test_version_data_ironic_microversions(self):
+        """Validate detection of Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}],
+                'version': '1.40',
+                'min_version': '1.10',
+            },
+            # Keep headers so we can verify that body trumps headers
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        self.assertEqual(
+            [
+                {
+                    'collection': None,
+                    'version': (1, 0),
+                    'url': ironic_url,
+                    'status': discover.Status.CURRENT,
+                    'raw_status': discover.Status.CURRENT,
+                    'min_microversion': (1, 10),
+                    'max_microversion': (1, 40),
+                    'next_min_version': None,
+                    'not_before': None,
+                },
+            ],
+            discover.Discover(self.session, ironic_url).version_data())
+
+    def test_version_data_legacy_ironic_microversions(self):
+        """Validate detection of legacy Ironic microversion ranges."""
+        ironic_url = 'https://bare-metal.example.com/v1/'
+        self.requests_mock.get(
+            ironic_url, status_code=200,
+            json={
+                'id': 'v1',
+                'links': [{
+                    "href": ironic_url,
+                    "rel": "self"}]},
+            headers={
+                'X-OpenStack-Ironic-API-Minimum-Version': '1.3',
+                'X-OpenStack-Ironic-API-Maximum-Version': '1.21',
+            })
+
+        self.assertEqual(
+            [
+                {
+                    'collection': None,
+                    'version': (1, 0),
+                    'url': ironic_url,
+                    'status': discover.Status.CURRENT,
+                    'raw_status': discover.Status.CURRENT,
+                    'min_microversion': (1, 3),
+                    'max_microversion': (1, 21),
+                    'next_min_version': None,
+                    'not_before': None,
+                },
+            ],
+            discover.Discover(self.session, ironic_url).version_data())
+
     def test_version_data_microversions(self):
         """Validate [min_|max_]version conversion to {min|max}_microversion."""
         def setup_mock(versions_in):
@@ -476,7 +657,7 @@ class VersionDataTests(utils.TestCase):
                 "versions": [
                     dict(
                         {
-                            "status": "CURRENT",
+                            "status": discover.Status.CURRENT,
                             "id": "v2.2",
                             "links": [
                                 {
@@ -502,7 +683,8 @@ class VersionDataTests(utils.TestCase):
                             'collection': None,
                             'version': (2, 2),
                             'url': V3_URL,
-                            'raw_status': 'CURRENT',
+                            'status': discover.Status.CURRENT,
+                            'raw_status': discover.Status.CURRENT,
                         },
                         **versions_out
                     )
@@ -556,6 +738,92 @@ class VersionDataTests(utils.TestCase):
         # Badly-formatted next_min_version
         test_exc({'next_min_version': 'bogus', 'not_before': '2019-07-01'})
 
+    def test_endpoint_data_noauth_discover(self):
+        mock = self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+        plugin = noauth.NoAuth()
+        data = plugin.get_endpoint_data(self.session, endpoint_override=V3_URL)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(
+            plugin.get_api_major_version(
+                self.session, endpoint_override=V3_URL),
+            (3, 0))
+        self.assertEqual(
+            plugin.get_endpoint(self.session, endpoint_override=V3_URL),
+            V3_URL)
+
+        self.assertTrue(mock.called_once)
+
+    def test_endpoint_data_noauth_no_discover(self):
+        plugin = noauth.NoAuth()
+        data = plugin.get_endpoint_data(
+            self.session, endpoint_override=V3_URL, discover_versions=False)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(
+            plugin.get_api_major_version(
+                self.session, endpoint_override=V3_URL),
+            (3, 0))
+        self.assertEqual(
+            plugin.get_endpoint(self.session, endpoint_override=V3_URL),
+            V3_URL)
+
+    def test_endpoint_data_noauth_adapter(self):
+        mock = self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+
+        client = adapter.Adapter(
+            session.Session(noauth.NoAuth()),
+            endpoint_override=V3_URL)
+        data = client.get_endpoint_data()
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(client.get_api_major_version(), (3, 0))
+        self.assertEqual(client.get_endpoint(), V3_URL)
+
+        self.assertTrue(mock.called_once)
+
+    def test_endpoint_data_token_endpoint_discover(self):
+        mock = self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+        plugin = token_endpoint.Token(endpoint=V3_URL, token='bogus')
+        data = plugin.get_endpoint_data(self.session)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)
+
+        self.assertTrue(mock.called_once)
+
+    def test_endpoint_data_token_endpoint_no_discover(self):
+        plugin = token_endpoint.Token(endpoint=V3_URL, token='bogus')
+        data = plugin.get_endpoint_data(self.session, discover_versions=False)
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(plugin.get_api_major_version(self.session), (3, 0))
+        self.assertEqual(plugin.get_endpoint(self.session), V3_URL)
+
+    def test_endpoint_data_token_endpoint_adapter(self):
+        mock = self.requests_mock.get(
+            V3_URL, status_code=200, json=V3_VERSION_ENTRY)
+        plugin = token_endpoint.Token(endpoint=V3_URL, token='bogus')
+
+        client = adapter.Adapter(session.Session(plugin))
+        data = client.get_endpoint_data()
+
+        self.assertEqual(data.api_version, (3, 0))
+        self.assertEqual(data.url, V3_URL)
+        self.assertEqual(client.get_api_major_version(), (3, 0))
+        self.assertEqual(client.get_endpoint(), V3_URL)
+
+        self.assertTrue(mock.called_once)
+
     def test_data_for_url(self):
         mock = self.requests_mock.get(V3_URL,
                                       status_code=200,
@@ -609,30 +877,34 @@ class VersionDataTests(utils.TestCase):
             self.assertIn(v['version'], ((2, 0), (3, 0)))
             self.assertEqual(v['raw_status'], 'stable')
 
-        for version in (disc.data_for('v3.0'),
-                        disc.data_for('3.latest'),
-                        disc.data_for('latest'),
-                        disc.versioned_data_for(
-                            min_version='v3.0', max_version='v3.latest'),
-                        disc.versioned_data_for(min_version='3'),
-                        disc.versioned_data_for(min_version='3.latest'),
-                        disc.versioned_data_for(min_version='latest'),
-                        disc.versioned_data_for(min_version='3.latest',
-                                                max_version='latest'),
-                        disc.versioned_data_for(min_version='latest',
-                                                max_version='latest'),
-                        disc.versioned_data_for(min_version=2),
-                        disc.versioned_data_for(min_version='2.latest')):
+        valid_v3_versions = (
+            disc.data_for('v3.0'),
+            disc.data_for('3.latest'),
+            disc.data_for('latest'),
+            disc.versioned_data_for(min_version='v3.0',
+                                    max_version='v3.latest'),
+            disc.versioned_data_for(min_version='3'),
+            disc.versioned_data_for(min_version='3.latest'),
+            disc.versioned_data_for(min_version='latest'),
+            disc.versioned_data_for(min_version='3.latest',
+                                    max_version='latest'),
+            disc.versioned_data_for(min_version='latest',
+                                    max_version='latest'),
+            disc.versioned_data_for(min_version=2),
+            disc.versioned_data_for(min_version='2.latest'))
+        for version in valid_v3_versions:
             self.assertEqual((3, 0), version['version'])
             self.assertEqual('stable', version['raw_status'])
             self.assertEqual(V3_URL, version['url'])
 
-        for version in (disc.data_for(2),
-                        disc.data_for('2.latest'),
-                        disc.versioned_data_for(
-                            min_version=2, max_version=(2, discover.LATEST)),
-                        disc.versioned_data_for(
-                            min_version='2.latest', max_version='2.latest')):
+        valid_v2_versions = (
+            disc.data_for(2),
+            disc.data_for('2.latest'),
+            disc.versioned_data_for(min_version=2,
+                                    max_version=(2, discover.LATEST)),
+            disc.versioned_data_for(min_version='2.latest',
+                                    max_version='2.latest'))
+        for version in valid_v2_versions:
             self.assertEqual((2, 0), version['version'])
             self.assertEqual('stable', version['raw_status'])
             self.assertEqual(V2_URL, version['url'])
@@ -661,7 +933,7 @@ class VersionDataTests(utils.TestCase):
         self.assertEqual(3, len(raw_data))
 
         for v in raw_data:
-            self.assertEqual(v['status'], 'CURRENT')
+            self.assertEqual(v['status'], discover.Status.CURRENT)
             if v['id'] == 'v1.0':
                 self.assertEqual(v['updated'], '2012-01-04T11:33:21Z')
             elif v['id'] == 'v2.0':
@@ -684,7 +956,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (1, 0),
                 'url': v1_url,
-                'raw_status': 'CURRENT',
+                'status': discover.Status.CURRENT,
+                'raw_status': discover.Status.CURRENT,
             },
             {
                 'collection': None,
@@ -694,7 +967,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (2, 0),
                 'url': v2_url,
-                'raw_status': 'CURRENT',
+                'status': discover.Status.CURRENT,
+                'raw_status': discover.Status.CURRENT,
             },
             {
                 'collection': BASE_URL,
@@ -704,7 +978,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': u'2019-12-31',
                 'version': (3, 0),
                 'url': v3_url,
-                'raw_status': 'CURRENT',
+                'status': discover.Status.CURRENT,
+                'raw_status': discover.Status.CURRENT,
             },
         ])
 
@@ -712,7 +987,7 @@ class VersionDataTests(utils.TestCase):
                         disc.versioned_data_for(min_version='v2.0',
                                                 max_version='v2.latest')):
             self.assertEqual((2, 0), version['version'])
-            self.assertEqual('CURRENT', version['raw_status'])
+            self.assertEqual(discover.Status.CURRENT, version['raw_status'])
             self.assertEqual(v2_url, version['url'])
 
         for version in (disc.data_for(1),
@@ -720,7 +995,7 @@ class VersionDataTests(utils.TestCase):
                             min_version=(1,),
                             max_version=(1, discover.LATEST))):
             self.assertEqual((1, 0), version['version'])
-            self.assertEqual('CURRENT', version['raw_status'])
+            self.assertEqual(discover.Status.CURRENT, version['raw_status'])
             self.assertEqual(v1_url, version['url'])
 
         self.assertIsNone(disc.url_for('v4'))
@@ -751,9 +1026,9 @@ class VersionDataTests(utils.TestCase):
 
         for v in raw_data:
             if v['id'] in ('v2.2', 'v1.1'):
-                self.assertEqual(v['status'], 'CURRENT')
+                self.assertEqual(v['status'], discover.Status.CURRENT)
             elif v['id'] in ('v2.1', 'v2.0', 'v1.0'):
-                self.assertEqual(v['status'], 'SUPPORTED')
+                self.assertEqual(v['status'], discover.Status.SUPPORTED)
             else:
                 self.fail("Invalid version found")
 
@@ -769,7 +1044,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (1, 0),
                 'url': v1_url,
-                'raw_status': 'SUPPORTED',
+                'status': discover.Status.SUPPORTED,
+                'raw_status': discover.Status.SUPPORTED,
             },
             {
                 'collection': None,
@@ -779,7 +1055,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (1, 1),
                 'url': v1_url,
-                'raw_status': 'CURRENT',
+                'status': discover.Status.CURRENT,
+                'raw_status': discover.Status.CURRENT,
             },
             {
                 'collection': None,
@@ -789,7 +1066,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (2, 0),
                 'url': v2_url,
-                'raw_status': 'SUPPORTED',
+                'status': discover.Status.SUPPORTED,
+                'raw_status': discover.Status.SUPPORTED,
             },
             {
                 'collection': None,
@@ -799,7 +1077,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (2, 1),
                 'url': v2_url,
-                'raw_status': 'SUPPORTED',
+                'status': discover.Status.SUPPORTED,
+                'raw_status': discover.Status.SUPPORTED,
             },
             {
                 'collection': None,
@@ -809,7 +1088,8 @@ class VersionDataTests(utils.TestCase):
                 'not_before': None,
                 'version': (2, 2),
                 'url': v2_url,
-                'raw_status': 'CURRENT',
+                'status': discover.Status.CURRENT,
+                'raw_status': discover.Status.CURRENT,
             },
         ])
 
@@ -819,7 +1099,9 @@ class VersionDataTests(utils.TestCase):
                                 min_version=ver,
                                 max_version=(2, discover.LATEST))):
                 self.assertEqual((2, 2), version['version'])
-                self.assertEqual('CURRENT', version['raw_status'])
+                self.assertEqual(
+                    discover.Status.CURRENT, version['raw_status']
+                )
                 self.assertEqual(v2_url, version['url'])
                 self.assertEqual(v2_url, disc.url_for(ver))
 
@@ -829,7 +1111,9 @@ class VersionDataTests(utils.TestCase):
                                 min_version=ver,
                                 max_version=(1, discover.LATEST))):
                 self.assertEqual((1, 1), version['version'])
-                self.assertEqual('CURRENT', version['raw_status'])
+                self.assertEqual(
+                    discover.Status.CURRENT, version['raw_status']
+                )
                 self.assertEqual(v1_url, version['url'])
                 self.assertEqual(v1_url, disc.url_for(ver))
 
@@ -949,3 +1233,20 @@ class EndpointDataTests(utils.TestCase):
         mock_get_disc.assert_has_calls(
             [mock.call('sess', url, cache='cache', authenticated=False)
              for url in ('url1', 'url2', 'url3')])
+
+    def test_endpoint_data_str(self):
+        """Validate EndpointData.__str__."""
+        # Populate a few fields to make sure they come through.
+        epd = discover.EndpointData(catalog_url='abc', service_type='123',
+                                    api_version=(2, 3))
+        exp = (
+            'EndpointData{api_version=(2, 3), catalog_url=abc,'
+            ' endpoint_id=None, interface=None, major_version=None,'
+            ' max_microversion=None, min_microversion=None,'
+            ' next_min_version=None, not_before=None, raw_endpoint=None,'
+            ' region_name=None, service_id=None, service_name=None,'
+            ' service_type=123, service_url=None, url=abc}')
+        # Works with str()
+        self.assertEqual(exp, str(epd))
+        # Works with implicit stringification
+        self.assertEqual(exp, "%s" % epd)
